#version 430 core
/*
* Copyright 2019 Vienna University of Technology.
* Institute of Computer Graphics and Algorithms.
* This file is part of the ECG Lab Framework and must not be redistributed.
*/

in VertexData {
	vec3 position_world; //WorldPos
	vec3 normal_world;		//Normal
	vec2 uv; //TexCoords
} vert;

out vec4 color; //FragColor

uniform vec3 camera_world; //camPos
uniform vec2 physicsCoefficients;
float roughness = physicsCoefficients.x;
float metallic = physicsCoefficients.y;
uniform vec3 materialCoefficients; // x = ambient, y = diffuse, z = specular 
uniform float specularAlpha;
uniform float illuminationMultiplier;
uniform sampler2D diffuseTexture;

vec3 Lo = vec3(0.0);
const float PI = 3.14159265358979323846264338327950288;

uniform struct DirectionalLight {
	vec3 color;
	vec3 direction;
} dirL;

uniform struct PointLight {
	vec3 color;
	vec3 position;
	vec3 attenuation;
} pointL;


//https://learnopengl.com/PBR/Lighting
vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
    return F0 + (1.0 - F0) * pow(max(1.0 - cosTheta, 0.0), 5.0);
}

float DistributionGGX(vec3 N, vec3 H, float roughness)
{
    float a      = roughness*roughness;
    float a2     = a*a;
    float NdotH  = max(dot(N, H), 0.0);
    float NdotH2 = NdotH*NdotH;
	
    float num   = a2;
    float denom = (NdotH2 * (a2 - 1.0) + 1.0);
    denom = PI * denom * denom;
	
    return num / denom;
}

float GeometrySchlickGGX(float NdotV, float roughness)
{
    float r = (roughness + 1.0);
    float k = (r*r) / 8.0;

    float num   = NdotV;
    float denom = NdotV * (1.0 - k) + k;
	
    return num / denom;
}
float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness)
{
    float NdotV = max(dot(N, V), 0.0);
    float NdotL = max(dot(N, L), 0.0);
    float ggx2  = GeometrySchlickGGX(NdotV, roughness);
    float ggx1  = GeometrySchlickGGX(NdotL, roughness);
	
    return ggx1 * ggx2;
}

vec3 pbs(vec3 n, vec3 l, vec3 v, vec3 diffuseC, float diffuseF, vec3 specularC, float specularF, float alpha, bool attenuate, vec3 attenuation) {
	float d = length(l);
	l = normalize(l);
	float att = 1.0;
	
	//BEGIN PBS
	vec3 h = normalize(v + l);
	vec3 F0 = vec3(0.04); 
	vec3 albedo = diffuseF * diffuseC * max(0, dot(n, l));
	F0  = mix(F0, albedo, metallic);
	vec3 F  = fresnelSchlick(max(dot(h, v), 0.0), F0);
	float NDF = DistributionGGX(n, h, roughness);       
	float G   = GeometrySmith(n, v, l, roughness); 
	vec3 numerator    = NDF * G * F;
	float denominator = 4.0 * max(dot(n, v), 0.0) * max(dot(n, l), 0.0);
	vec3 specular     = numerator / max(denominator, 0.001); 
	vec3 kS = F;
	vec3 kD = vec3(1.0) - kS;
	kD *= 1.0 - metallic;
	float NdotL = max(dot(n, l), 0.0);  

	if(attenuate) att = 1.0f / (attenuation.x + d * attenuation.y + d * d * attenuation.z);

	vec3  radiance    = specularC * att * max(dot(h, v), 0.0);
    Lo += (kD * albedo / PI + specular) * radiance * NdotL;
	Lo = Lo / (Lo + vec3(1.0));
	Lo = pow(Lo, vec3(1.0/1.5));
	return Lo;
	
	
	if(attenuate) att = 1.0f / (attenuation.x + d * attenuation.y + d * d * attenuation.z);
	vec3 r = reflect(-l, n);
	return (diffuseF * diffuseC * max(0, dot(n, l)) + specularF * specularC * pow(max(0, dot(r, v)), alpha)) * att; 
	
}


void main() {	
	vec3 n = normalize(vert.normal_world);
	vec3 v = normalize(camera_world - vert.position_world);


	
	vec3 texColor = texture(diffuseTexture, vert.uv).rgb;
	color = vec4(illuminationMultiplier * texColor * materialCoefficients.x, 1); // ambient
	//color=vec4(0,0,0,1);
	
	// add directional light contribution
	color.rgb += pbs(n, -dirL.direction, v, dirL.color * texColor, materialCoefficients.y, dirL.color, materialCoefficients.z, specularAlpha, false, vec3(0));
			
	// add point light contribution
	color.rgb += pbs(n, pointL.position - vert.position_world, v, pointL.color * texColor, materialCoefficients.y, pointL.color, materialCoefficients.z, specularAlpha, true, pointL.attenuation);

}

